"""Define the DIMSE Message classes."""

from io import BytesIO
import logging
from math import ceil
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterator, Optional, TYPE_CHECKING, cast, Union, Tuple

from pydicom.dataset import Dataset
from pydicom.filewriter import write_file_meta_info
from pydicom.tag import Tag
from pydicom.uid import UID

from pynetdicom import _config
from pynetdicom.dimse_primitives import (
    C_STORE,
    C_FIND,
    C_GET,
    C_MOVE,
    C_ECHO,
    C_CANCEL,
    N_EVENT_REPORT,
    N_GET,
    N_SET,
    N_ACTION,
    N_CREATE,
    N_DELETE,
    DimsePrimitiveType,
)
from pynetdicom.dsutils import encode, decode, create_file_meta
from pynetdicom.pdu_primitives import P_DATA

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.association import Association
    from pynetdicom.dimse_primitives import NTF


LOGGER = logging.getLogger(__name__)


# PS3.7 Section 9.3
_COMMAND_SET_KEYWORDS = {
    "C-ECHO-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
    ),
    "C-ECHO-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "ErrorComment",
    ),
    "C-STORE-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "Priority",
        "CommandDataSetType",
        "AffectedSOPInstanceUID",
        "MoveOriginatorApplicationEntityTitle",
        "MoveOriginatorMessageID",
    ),
    "C-STORE-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "OffendingElement",
        "ErrorComment",
    ),
    "C-FIND-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "Priority",
        "CommandDataSetType",
    ),
    "C-FIND-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "OffendingElement",
        "ErrorComment",
    ),
    "C-CANCEL-RQ": (
        "CommandGroupLength",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
    ),
    "C-GET-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "Priority",
        "CommandDataSetType",
    ),
    "C-GET-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "NumberOfRemainingSuboperations",
        "NumberOfCompletedSuboperations",
        "NumberOfFailedSuboperations",
        "NumberOfWarningSuboperations",
        "OffendingElement",
        "ErrorComment",
    ),
    "C-MOVE-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "Priority",
        "CommandDataSetType",
        "MoveDestination",
    ),
    "C-MOVE-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "NumberOfRemainingSuboperations",
        "NumberOfCompletedSuboperations",
        "NumberOfFailedSuboperations",
        "NumberOfWarningSuboperations",
        "OffendingElement",
        "ErrorComment",
    ),
    "N-EVENT-REPORT-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
        "AffectedSOPInstanceUID",
        "EventTypeID",
    ),
    "N-EVENT-REPORT-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "EventTypeID",
        "ErrorID",
        "ErrorComment",
    ),
    "N-GET-RQ": (
        "CommandGroupLength",
        "RequestedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
        "RequestedSOPInstanceUID",
        "AttributeIdentifierList",
    ),
    "N-GET-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "AttributeIdentifierList",
        "ErrorComment",
        "ErrorID",
    ),
    "N-SET-RQ": (
        "CommandGroupLength",
        "RequestedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
        "RequestedSOPInstanceUID",
    ),
    "N-SET-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "AttributeIdentifierList",
        "ErrorComment",
        "ErrorID",
    ),
    "N-ACTION-RQ": (
        "CommandGroupLength",
        "RequestedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
        "RequestedSOPInstanceUID",
        "ActionTypeID",
    ),
    "N-ACTION-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "ActionTypeID",
        "ErrorID",
        "ErrorComment",
    ),
    "N-CREATE-RQ": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
        "AffectedSOPInstanceUID",
    ),
    "N-CREATE-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "ErrorID",
        "ErrorComment",
    ),
    "N-DELETE-RQ": (
        "CommandGroupLength",
        "RequestedSOPClassUID",
        "CommandField",
        "MessageID",
        "CommandDataSetType",
        "RequestedSOPInstanceUID",
    ),
    "N-DELETE-RSP": (
        "CommandGroupLength",
        "AffectedSOPClassUID",
        "CommandField",
        "MessageIDBeingRespondedTo",
        "CommandDataSetType",
        "Status",
        "AffectedSOPInstanceUID",
        "ErrorComment",
        "ErrorID",
    ),
}

# Used with DIMSEMessage.message_to_primitive
_MSG_TO_PRIMITVE = {
    "C_ECHO": C_ECHO,
    "C_STORE": C_STORE,
    "C_FIND": C_FIND,
    "C_GET": C_GET,
    "C_MOVE": C_MOVE,
    "C_CANCEL": C_CANCEL,
    "N_EVENT_REPORT": N_EVENT_REPORT,
    "N_GET": N_GET,
    "N_SET": N_SET,
    "N_ACTION": N_ACTION,
    "N_CREATE": N_CREATE,
    "N_DELETE": N_DELETE,
}

_MULTIVALUE_TAGS = [Tag("OffendingElement"), Tag("AttributeIdentifierList")]


class DIMSEMessage:
    """Represents a DIMSE Message.

    Information is communicated across the DICOM network interface in a DICOM
    Message. A Message is composed of a Command Set followed by a
    conditional Data Set. The Command Set is used to indicate the
    operations/notifications to be performed on or with the Data Set (see
    PS3.7 6.2-3).

    ::

                primitive_to_message()       encode_msg()
        +-------------+   --->    +-----------+   --->   +-------------+
        |    DIMSE    |           |   DIMSE   |          |   P-DATA    |
        |  primitive  |           |  Message  |          |  primitive  |
        +-------------+   <---    +-----------+   <---   +-------------+
                message_to_primitive()       decode_msg()


    **Command Set**

    A Command Set is constructed of Command Elements. Command Elements contain
    the encoded values for each field of the Command Set per the semantics
    specified in the DIMSE protocol (PS3.7 9.2 and 10.2). Each Command Element
    is composed of an explicit Tag, Value Length and a Value field. The
    encoding of the Command Set shall be *Little Endian Implicit VR*.

    **Message Types**

    The following message types are available: C_STORE_RQ, C_STORE_RSP,
    C_ECHO_RQ, C_ECHO_RSP, C_FIND_RQ, C_FIND_RSP, C_GET_RQ, C_GET_RSP,
    C_MOVE_RQ, C_MOVE_RSP, C_CANCEL_RQ, N_EVENT_REPORT_RQ, N_EVENT_REPORT_RSP,
    N_SET_RQ, N_SET_RSP, N_GET_RQ, N_GET_RSP, N_ACTION_RQ, N_ACTION_RSP,
    N_CREATE_RQ, N_CREATE_RSP, N_DELETE_RQ, N_DELETE_RSP.

    Attributes
    ----------
    command_set : pydicom.dataset.Dataset
        The message Command Set information (see PS3.7 6.3).
    context_id : int
        The presentation context ID.
    data_set : io.BytesIO
        The encoded message Data Set (see PS3.7 6.3).
    encoded_command_set : io.BytesIO
        During decoding of an incoming P-DATA primitive this stores the
        encoded Command Set data from the fragments.
    """

    def __init__(self) -> None:
        """Create a new DIMSE Message."""
        self.context_id: Optional[int] = None

        # Required to save command set data from multiple fragments
        self.encoded_command_set = BytesIO()
        self.data_set: Optional[BytesIO] = BytesIO()
        self.command_set = Dataset()

        # If reading the dataset in chunks this will be a tuple:
        #   (its file path, a byte offset to the start of the dataset)
        # If writing the dataset in chunks this will be a Path:
        #   its file path
        self._data_set_path: Optional[Union[Path, Tuple[Path, int]]] = None
        # If writing the dataset in chunks this will be a NamedTemporaryFile:
        #   the file object backing its file path
        self._data_set_file: Optional["NTF"] = None

        cls_name = self.__class__.__name__
        if cls_name == "DIMSEMessage":
            return

        # Set the command set attributes for the subclasses
        for keyword in _COMMAND_SET_KEYWORDS[cls_name.replace("_", "-")]:
            setattr(self.command_set, keyword, None)

    def decode_msg(
        self, primitive: P_DATA, assoc: Optional["Association"] = None
    ) -> bool:
        """Converts P-DATA primitives into a ``DIMSEMessage`` sub-class.

        Decodes the data from the P-DATA service primitive (which
        may contain the results of one or more P-DATA-TF PDUs) into the
        :attr:`~DIMSEMessage.command_set` and :attr:`~DIMSEMessage.data_set`
        attributes. Also sets the :attr:`~DIMSEMessage.context_id` and
        :attr:`~DIMSEMessage.encoded_command_set` attributes of the
        ``DIMSEMessage`` sub-class object.

        Parameters
        ----------
        primitive : pdu_primitives.P_DATA
            The P-DATA service primitive to be decoded into a DIMSE message.
        assoc : association.Association, optional
            The association processing the message. This is required when:

            * :attr:`~pynetdicom._config.STORE_RECV_CHUNKED_DATASET` is
              ``True``
            * The P-DATA primitive contains part of a C-STORE-RQ message

            In this case the association is consulted for its accepted
            transfer syntax, which is included in the File Meta Information
            of the stored dataset.

        Returns
        -------
        bool
            ``True`` when the DIMSE message is completely decoded, ``False``
            otherwise.

        References
        ----------

        * DICOM Standard, Part 8, :dcm:`Annex E<part08/chapter_E.html>`
        """
        # Make sure this is a P-DATA primitive
        if primitive.__class__ != P_DATA or primitive is None:
            return False

        for context_id, data in primitive.presentation_data_value_list:
            # The first byte of the P-DATA is the Message Control Header
            #   See Part 8, Annex E.2
            # The standard says that only the significant bits (ie the last
            #   two) should be checked
            # xxxxxx00 - Message Dataset information, not the last fragment
            # xxxxxx01 - Command information, not the last fragment
            # xxxxxx10 - Message Dataset information, the last fragment
            # xxxxxx11 - Command information, the last fragment
            control_header_byte = data[0]

            # COMMAND SET
            # P-DATA fragment contains Command Set information
            #   (control_header_byte is xxxxxx01 or xxxxxx11)
            if control_header_byte & 1:
                # The command set may be spread out over a number
                #   of fragments and P-DATA primitives and we need to remember
                #   the elements from previous fragments, hence the
                #   encoded_command_set class attribute
                # This adds all the command set data to the class object
                self.encoded_command_set.write(data[1:])

                # The final command set fragment (xxxxxx11) has been added
                #   so decode the command set
                if control_header_byte & 2:
                    # Presentation Context ID
                    #   Set this now as must only be one final command set
                    #   fragment and command set must always be present
                    self.context_id = context_id

                    # Command Set is always encoded Implicit VR Little Endian
                    #   decode(dataset, is_implicit_VR, is_little_endian)
                    # pylint: disable=attribute-defined-outside-init
                    self.command_set = decode(self.encoded_command_set, True, True)

                    # Determine which DIMSE Message class to use
                    self.__class__ = _MESSAGE_TYPES[
                        cast(int, self.command_set.CommandField)
                    ][1]

                    # Determine if a Data Set is present by checking for
                    #   (0000, 0800) CommandDataSetType US 1. If the value is
                    #   0x0101 no dataset present, otherwise one is.
                    if self.command_set.CommandDataSetType == 0x0101:
                        # By returning True we're indicating that the message
                        #   has been completely decoded
                        return True

                    # Data Set is present
                    if _config.STORE_RECV_CHUNKED_DATASET and isinstance(
                        self, C_STORE_RQ
                    ):
                        # delete=False is a workaround for Windows
                        # Setting delete=True prevents us from re-opening
                        # the file after it is opened by NamedTemporaryFile
                        # below.
                        self._data_set_file = cast(
                            "NTF",
                            NamedTemporaryFile(delete=False, mode="wb", suffix=".dcm"),
                        )
                        self._data_set_path = Path(self._data_set_file.name)
                        # Write the File Meta
                        self._data_set_file.write(b"\x00" * 128)
                        self._data_set_file.write(b"DICM")

                        cs = self.command_set
                        cx = cast("Association", assoc)._accepted_cx[context_id]
                        sop_class = cast(UID, cs.AffectedSOPClassUID)
                        sop_instance = cast(UID, cs.AffectedSOPInstanceUID)
                        write_file_meta_info(
                            self._data_set_file,  # type: ignore
                            create_file_meta(
                                sop_class_uid=sop_class,
                                sop_instance_uid=sop_instance,
                                transfer_syntax=cx.transfer_syntax[0],
                            ),
                        )

            # DATA SET
            # P-DATA fragment contains Data Set information
            #   (control_header_byte is xxxxxx00 or xxxxxx10)
            else:
                # As with the command set, the data set may be spread over
                #   a number of fragments in each P-DATA primitive and a
                #   number of P-DATA primitives.
                if self._data_set_file:
                    self._data_set_file.write(data[1:])
                    self._data_set_file.file.flush()
                else:
                    cast(BytesIO, self.data_set).write(data[1:])

                # The final data set fragment (xxxxxx10) has been added
                if control_header_byte & 2 != 0:
                    # By returning True we're indicating that the message
                    #   has been completely decoded
                    return True

        # We return False to indicate that the message isn't yet fully decoded
        return False

    def encode_msg(self, context_id: int, max_pdu_length: int) -> Iterator[P_DATA]:
        """Yield P-DATA primitives for the current DIMSE Message.

        **Encoding**

        The encoding of the Command Set shall be *Little Endian Implicit VR*,
        while the *Data Set* will be encoded as per the agreed presentation
        context.

        A P-DATA request's PDV List parameter shall contain one or more PDVs.
        Each PDV is wholly contained in a given P-DATA request and doesn't
        span across several P-DATA request primitives.

        The fragmentation of any message results in a series of PDVs that shall
        be sent, on a given association, by a corresponding series of P-DATA
        requests preserving the ordering of the fragments of any message.
        No fragments of any other messages shall be sent until all fragments
        of the current message have been sent.

        Parameters
        ----------
        context_id : int
            The *ID* of the agreed presentation context.
        max_pdu_length : int
            The maximum PDV length (in bytes).

        Yields
        ------
        pdu_primitives.P_DATA
            The current DIMSE message as one or more P-DATA service
            primitives.

        References
        ----------

        * DICOM Standard, Part 7,
          :dcm:`Section 6.3.1<part07/sect_6.3.html#sect_6.3.1>`
        * DICOM Standard, Part 8, :dcm:`Annex E<part08/chapter_E.html>`
        """
        self.context_id = context_id

        # The Command Set is always Little Endian Implicit VR (PS3.7 6.3.1)
        #   encode(dataset, is_implicit_VR, is_little_endian)
        encoded_command_set = cast(bytes, encode(self.command_set, True, True))

        # COMMAND SET (always)
        # Split the command set into fragments with maximum size max_pdu_length
        if max_pdu_length == 0:
            nr_fragments = 1
        else:
            nr_fragments = ceil(len(encoded_command_set) / (max_pdu_length - 6))

        cmd_fragments = self._generate_pdv_fragments(
            encoded_command_set, max_pdu_length
        )

        # First to (n - 1)th command data fragment - bits xxxxxx01
        for ii in range(int(nr_fragments - 1)):
            pdata = P_DATA()
            pdata.presentation_data_value_list.append(
                (context_id, b"\x01" + next(cmd_fragments))
            )
            yield pdata

        # Last command data fragment - bits xxxxxx11
        pdata = P_DATA()
        pdata.presentation_data_value_list.append(
            (context_id, b"\x03" + next(cmd_fragments))
        )
        yield pdata

        # DATASET (if available)
        #   Check that the Data Set is not empty
        if self.data_set is not None:
            encoded_data_set = self.data_set.getvalue()
            if encoded_data_set:
                # Split the data set into fragments with maximum
                #   size max_pdu_length
                if max_pdu_length == 0:
                    nr_fragments = 1
                else:
                    nr_fragments = ceil(len(encoded_data_set) / (max_pdu_length - 6))

                ds_fragments = self._generate_pdv_fragments(
                    encoded_data_set, max_pdu_length
                )

                # First to (n - 1)th dataset fragment - bits xxxxxx00
                for ii in range(int(nr_fragments - 1)):
                    pdata = P_DATA()
                    pdata.presentation_data_value_list.append(
                        (context_id, b"\x00" + next(ds_fragments))
                    )
                    yield pdata

                # Last dataset fragment - bits xxxxxx10
                pdata = P_DATA()
                pdata.presentation_data_value_list.append(
                    (context_id, b"\x02" + next(ds_fragments))
                )
                yield pdata
        elif self._data_set_path is not None:
            # Read and send encoded dataset from file
            # Buffer size determined by io.DEFAULT_BUFFER_SIZE
            self._data_set_path = cast(Tuple[Path, int], self._data_set_path)
            with open(self._data_set_path[0], "rb") as f:
                end = f.seek(0, 2)
                length = end - f.seek(self._data_set_path[1])

                if max_pdu_length == 0:
                    nr_fragments = 1
                    max_pdu_length = length + 6
                else:
                    nr_fragments = ceil(length / (max_pdu_length - 6))

                # First to (n - 1)th dataset fragment - bits xxxxxx00
                for ii in range(int(nr_fragments - 1)):
                    pdata = P_DATA()
                    pdata.presentation_data_value_list.append(
                        (context_id, b"\x00" + f.read(max_pdu_length - 6))
                    )
                    yield pdata

                # Last dataset fragment - bits xxxxxx10
                pdata = P_DATA()
                pdata.presentation_data_value_list.append(
                    (context_id, b"\x02" + f.read(max_pdu_length - 6))
                )
                yield pdata

    @staticmethod
    def _generate_pdv_fragments(
        bytestream: bytes, fragment_length: int
    ) -> Iterator[bytes]:
        """Fragment `bytestream` into chunks, each `fragment_length` long.

        Fragments bytestream data for use in PDVs.

        The Maximum Length Negotiation allows receivers to limit the size of
        the Presentation Data Values List parameters of each P-DATA indication.

        The Association requestor shall specify the maximum length in bytes for
        the PDV list parameter it is ready to receive in each P-DATA
        indication. The Association acceptor shall ensure in its fragmentation
        of the DICOM Messages that the list of PDVs included in each P-DATA
        request does not exceed this maximum length.

        Parameters
        ----------
        bytestream : bytes
            The data to be fragmented.
        fragment_length : int
            The maximum size of each fragment, a value of 0 is taken to mean
            the fragment is infinite. Cannot be between 1 and 7 as
            each Presentation Data Value Item is::

                1 - 4       | 5          | 6 ->
                Item length | Context ID | Presentation data value ->

        Yields
        ------
        fragment : bytes
            A `bytestream` fragment, with maximum length `fragment_length`, but
            may be smaller depending on the size of `bytestream`.

        References
        ----------
        DICOM Standard, Part 8, Annex D.1
        """
        if fragment_length == 0:
            yield bytestream
            return
        elif 0 < fragment_length < 7:
            raise ValueError("'fragment_length' cannot be between 1 and 7.")

        # Because the PDV item includes an extra 6 bytes of data at the start
        #   we need to decrease `fragment_length` by 6 bytes.
        fragment_length -= 6

        offset = 0
        no_pdv = ceil(len(bytestream) / fragment_length)
        for ii in range(int(no_pdv)):
            yield bytestream[offset : offset + fragment_length]
            offset += fragment_length

    def message_to_primitive(self) -> DimsePrimitiveType:
        """Convert the ``DIMSEMessage`` class to a DIMSE primitive.

        Returns
        -------
        DIMSEPrimitive sub-class
            One of the DIMSE message primitives from
            :ref:`pynetdicom.dimse_primitives<api_dimse_primitives>` generated
            from the current ``DIMSEMessage`` sub-class object.
        """
        cls_type_name = self.__class__.__name__
        final_underscore = cls_type_name.rfind("_R")
        primitive = cast(
            DimsePrimitiveType, _MSG_TO_PRIMITVE[cls_type_name[:final_underscore]]()
        )

        # Command Set
        # For each parameter in the primitive, set the appropriate value
        #   from the Message's Command Set elements
        for elem in self.command_set:
            if hasattr(primitive, elem.keyword):
                value = elem.value
                if elem.VM > 1 and elem.tag not in _MULTIVALUE_TAGS:
                    LOGGER.warning(
                        f"Non-conformant VM {elem.VM} for '{elem.keyword}', "
                        "taking the first value"
                    )
                    value = value[0]
                setattr(primitive, elem.keyword, value)

        # Datasets
        # Set the primitive's DataSet/Identifier/etc attribute
        try:
            dataset_keyword = _DATASET_KEYWORDS[cls_type_name]
            setattr(primitive, dataset_keyword, self.data_set)
        except KeyError:
            pass

        # Set the presentation context ID the message was set under
        primitive._context_id = self.context_id

        primitive._dataset_path = self._data_set_path
        primitive._dataset_file = self._data_set_file

        return primitive

    def primitive_to_message(self, primitive: DimsePrimitiveType) -> None:
        """Convert a DIMSE `primitive` to the current ``DIMSEMessage`` object.

        Parameters
        ----------
        DIMSEPrimitive sub-class
            A DIMSE message primitive from
            :ref:`pynetdicom.dimse_primitives<api_dimse_primitives>`
            to convert to the current ``DIMSEMessage`` object.
        """
        # pylint: disable=too-many-branches,too-many-statements
        cls_type_name = self.__class__.__name__.replace("_", "-")
        if cls_type_name not in _COMMAND_SET_KEYWORDS:
            raise ValueError(
                f"Can't convert primitive to message for unknown "
                f"DIMSE message type '{cls_type_name}'"
            )

        # Command Set
        # Convert the message command set to the primitive attributes
        for elem in self.command_set:
            # Use the element keyword as these should match the parameter
            #   names in the primitive
            if hasattr(primitive, elem.keyword):
                # If value hasn't been set for a parameter then delete
                #   the corresponding element
                attr = getattr(primitive, elem.keyword)
                if attr is not None:
                    elem.value = attr
                else:
                    del self.command_set[elem.tag]

        # There's a one-to-one relationship in the _MESSAGE_TYPES dict, so
        #   invert it for convenience
        rev_type = {vv[0]: kk for kk, vv in _MESSAGE_TYPES.items()}
        self.command_set.CommandField = rev_type[cls_type_name]

        # Data Set
        # Default to no Data Set
        self.data_set = BytesIO()
        self.command_set.CommandDataSetType = 0x0101

        try:
            # These message types *may* have a dataset
            dataset_keyword = _DATASET_KEYWORDS[self.__class__.__name__]
            self.data_set = getattr(primitive, dataset_keyword)
            if self.data_set:
                self.command_set.CommandDataSetType = 0x0001
        except KeyError:
            # The following message types never have a dataset
            # 'C_ECHO_RQ', 'C_ECHO_RSP', 'N_DELETE_RQ', 'C_STORE_RSP',
            # 'C_CANCEL_RQ', 'N_DELETE_RSP', 'C_FIND_RSP', 'N_GET_RQ'
            pass

        self._data_set_path = getattr(primitive, "_dataset_path", None)
        if self._data_set_path:
            self.command_set.CommandDataSetType = 0x0001

        # Set the Command Set length
        self._set_command_group_length()

    def _set_command_group_length(self) -> None:
        """Reset the Command Group Length element value.

        Once the `command_set` Dataset has been built and filled with
        values, this should be called to set the (Command Group Length* element
        value correctly.
        """
        # Remove CommandGroupLength to stop it messing up the length calc
        del self.command_set.CommandGroupLength

        # The Command Set is always Implicit VR Little Endian
        self.command_set.CommandGroupLength = len(
            cast(bytes, encode(self.command_set, True, True))
        )


# Create DIMSEMessage subclasses
class C_STORE_RQ(DIMSEMessage):
    pass


class C_STORE_RSP(DIMSEMessage):
    pass


class C_FIND_RQ(DIMSEMessage):
    pass


class C_FIND_RSP(DIMSEMessage):
    pass


class C_GET_RQ(DIMSEMessage):
    pass


class C_GET_RSP(DIMSEMessage):
    pass


class C_MOVE_RQ(DIMSEMessage):
    pass


class C_MOVE_RSP(DIMSEMessage):
    pass


class C_ECHO_RQ(DIMSEMessage):
    pass


class C_ECHO_RSP(DIMSEMessage):
    pass


class C_CANCEL_RQ(DIMSEMessage):
    pass


class N_EVENT_REPORT_RQ(DIMSEMessage):
    pass


class N_EVENT_REPORT_RSP(DIMSEMessage):
    pass


class N_GET_RQ(DIMSEMessage):
    pass


class N_GET_RSP(DIMSEMessage):
    pass


class N_SET_RQ(DIMSEMessage):
    pass


class N_SET_RSP(DIMSEMessage):
    pass


class N_ACTION_RQ(DIMSEMessage):
    pass


class N_ACTION_RSP(DIMSEMessage):
    pass


class N_CREATE_RQ(DIMSEMessage):
    pass


class N_CREATE_RSP(DIMSEMessage):
    pass


class N_DELETE_RQ(DIMSEMessage):
    pass


class N_DELETE_RSP(DIMSEMessage):
    pass


# Values from PS3.5
_MESSAGE_TYPES = {
    0x0001: ("C-STORE-RQ", C_STORE_RQ),
    0x8001: ("C-STORE-RSP", C_STORE_RSP),
    0x0020: ("C-FIND-RQ", C_FIND_RQ),
    0x8020: ("C-FIND-RSP", C_FIND_RSP),
    0x0010: ("C-GET-RQ", C_GET_RQ),
    0x8010: ("C-GET-RSP", C_GET_RSP),
    0x0021: ("C-MOVE-RQ", C_MOVE_RQ),
    0x8021: ("C-MOVE-RSP", C_MOVE_RSP),
    0x0030: ("C-ECHO-RQ", C_ECHO_RQ),
    0x8030: ("C-ECHO-RSP", C_ECHO_RSP),
    0x0FFF: ("C-CANCEL-RQ", C_CANCEL_RQ),
    0x0100: ("N-EVENT-REPORT-RQ", N_EVENT_REPORT_RQ),
    0x8100: ("N-EVENT-REPORT-RSP", N_EVENT_REPORT_RSP),
    0x0110: ("N-GET-RQ", N_GET_RQ),
    0x8110: ("N-GET-RSP", N_GET_RSP),
    0x0120: ("N-SET-RQ", N_SET_RQ),
    0x8120: ("N-SET-RSP", N_SET_RSP),
    0x0130: ("N-ACTION-RQ", N_ACTION_RQ),
    0x8130: ("N-ACTION-RSP", N_ACTION_RSP),
    0x0140: ("N-CREATE-RQ", N_CREATE_RQ),
    0x8140: ("N-CREATE-RSP", N_CREATE_RSP),
    0x0150: ("N-DELETE-RQ", N_DELETE_RQ),
    0x8150: ("N-DELETE-RSP", N_DELETE_RSP),
}

_DATASET_KEYWORDS = {
    "C_STORE_RQ": "DataSet",
    "C_FIND_RQ": "Identifier",
    "C_GET_RQ": "Identifier",
    "C_MOVE_RQ": "Identifier",
    "C_FIND_RSP": "Identifier",
    "C_GET_RSP": "Identifier",
    "C_MOVE_RSP": "Identifier",
    "N_EVENT_REPORT_RQ": "EventInformation",
    "N_EVENT_REPORT_RSP": "EventReply",
    "N_GET_RSP": "AttributeList",
    "N_SET_RSP": "AttributeList",
    "N_CREATE_RQ": "AttributeList",
    "N_CREATE_RSP": "AttributeList",
    "N_SET_RQ": "ModificationList",
    "N_ACTION_RQ": "ActionInformation",
    "N_ACTION_RSP": "ActionReply",
}



"""Tests for the ae module."""

import logging
import os
import signal
import threading
import time

import pytest

from pydicom import dcmread, config as PYD_CONFIG
from pydicom.dataset import Dataset
from pydicom.uid import UID, ImplicitVRLittleEndian

from pynetdicom import (
    AE,
    build_context,
    _config,
    debug_logger,
    DEFAULT_TRANSFER_SYNTAXES,
    evt,
    PYNETDICOM_IMPLEMENTATION_UID,
    PYNETDICOM_IMPLEMENTATION_VERSION,
    StoragePresentationContexts,
    VerificationPresentationContexts,
)
from pynetdicom.presentation import build_context
from pynetdicom.sop_class import RTImageStorage, Verification
from pynetdicom.transport import AssociationServer, RequestHandler


if hasattr(PYD_CONFIG, "settings"):
    PYD_CONFIG.settings.reading_validation_mode = 0


# debug_logger()


TEST_DS_DIR = os.path.join(os.path.dirname(__file__), "dicom_files")
DATASET = dcmread(os.path.join(TEST_DS_DIR, "RTImageStorage.dcm"))
COMP_DATASET = dcmread(os.path.join(TEST_DS_DIR, "MRImageStorage_JPG2000_Lossless.dcm"))


def test_blocking_handler():
    """Test binding events to the blocking AssociationServer."""
    ae = AE()
    ae.add_supported_context("1.2.840.10008.1.1")

    def handle_echo(event):
        return 0x0000

    handlers = [(evt.EVT_C_ECHO, handle_echo)]

    thread = threading.Thread(
        target=ae.start_server,
        args=(("localhost", 11112),),
        kwargs={"evt_handlers": handlers},
    )
    thread.daemon = True
    thread.start()

    time.sleep(0.1)

    ae.shutdown()


class TestMakeServer:
    """Tests for AE.make_server()"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_default_arguments(self):
        self.ae = ae = AE()
        ae.add_supported_context(Verification)

        server = ae.make_server(("localhost", 11112))
        assert isinstance(server, AssociationServer)

    def test_custom_request_handler(self):
        class MyRequestHandler(RequestHandler):
            pass

        self.ae = ae = AE()
        ae.add_supported_context(Verification)

        server = ae.make_server(("localhost", 11112), request_handler=MyRequestHandler)
        assert server.RequestHandlerClass is MyRequestHandler

    def test_aet_bytes_deprecation(self):
        """Test warning if using bytes to set an AE title."""
        self.ae = ae = AE()
        ae.add_supported_context(Verification)

        msg = (
            r"The use of bytes with 'ae_title' is deprecated, use an ASCII "
            r"str instead"
        )
        with pytest.warns(DeprecationWarning, match=msg):
            server = ae.start_server(
                ("localhost", 11112), block=False, ae_title=b"BADAE2"
            )
            assert server.ae_title == "BADAE2"
            server.shutdown()


class TestStartServer:
    """Tests for AE.start_server()"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_ae_title(self):
        """Test the `ae_title` keyword parameter."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.ae_title = "TESTAET"
        assert ae.ae_title == "TESTAET"

        ae.add_supported_context(Verification)
        server = ae.start_server(("localhost", 11112), block=False)
        assert server.ae_title == ae.ae_title

        server.shutdown()

        server = ae.start_server(("localhost", 11112), block=False, ae_title="MYAE")
        assert server.ae_title == "MYAE"
        ae.require_called_aet = True

        ae.add_requested_context(Verification)
        assoc = ae.associate("localhost", 11112, ae_title="MYAE")
        assert assoc.is_established
        assoc.release()
        assert assoc.is_released

        server.shutdown()

    def test_contexts(self):
        """Test the `contexts` keyword parameter."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.ae_title = "TESTAET"
        assert ae.ae_title == "TESTAET"

        cx = build_context(Verification)
        server = ae.start_server(("localhost", 11112), block=False, contexts=[cx])

        ae.add_requested_context(Verification)
        assoc = ae.associate("localhost", 11112, ae_title="MYAE")
        assert assoc.is_established
        assert assoc.accepted_contexts[0].abstract_syntax == Verification
        assoc.release()
        assert assoc.is_released

        server.shutdown()


class TestAEVerificationSCP:
    """Check verification SCP"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    @pytest.mark.skipif(os.name == "nt", reason="Kills pytest on windows")
    def test_start_server_keyboard_interrupt(self):
        """Test stopping the SCP with keyboard"""
        pid = os.getpid()

        def trigger_signal():
            time.sleep(0.1)
            os.kill(pid, signal.SIGINT)

        self.ae = ae = AE()
        ae.add_supported_context("1.2.3")
        thread = threading.Thread(target=trigger_signal)
        thread.daemon = True
        thread.start()

        ae.start_server(("localhost", 11112))

        ae.shutdown()

    def test_no_supported_contexts(self):
        """Test starting with no contexts raises"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        with pytest.raises(ValueError, match=r"No supported Presentation"):
            ae.start_server(("localhost", 11112))

    def test_new_scu_scp_warning(self):
        """Test that a warning is given if scu_role and scp_role bad."""
        ae = AE()
        ae.add_supported_context("1.2.3.4", scp_role=False)
        msg = r"The following presentation contexts have "
        with pytest.raises(ValueError, match=msg):
            ae.start_server(("localhost", 11112))

    def test_str_empty(self):
        """Test str output for default AE"""
        ae = AE()
        ae.__str__()


class TestAEPresentationSCU:
    """Tests for AE presentation contexts when running as an SCU"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_associate_context(self):
        """Test that AE.associate doesn't modify the supplied contexts"""
        # Test AE.requested_contexts
        self.ae = ae = AE()
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.requested_contexts = VerificationPresentationContexts
        ae.acse_timeout = 5
        ae.dimse_timeout = 5

        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assert ae.requested_contexts[0].context_id is None
        assert len(assoc.requestor.requested_contexts) == 1
        assert assoc.requestor.requested_contexts[0].abstract_syntax == (
            "1.2.840.10008.1.1"
        )
        assert assoc.requestor.requested_contexts[0].context_id == 1
        assoc.release()
        assert not assoc.is_established
        assert assoc.is_released

        # Test associate(contexts=...)
        ae.requested_contexts = []
        assoc = ae.associate(
            "localhost", 11112, contexts=VerificationPresentationContexts
        )
        assert assoc.is_established

        assert VerificationPresentationContexts[0].context_id is None
        assert len(assoc.requestor.requested_contexts) == 1
        assert assoc.requestor.requested_contexts[0].abstract_syntax == (
            "1.2.840.10008.1.1"
        )
        assert assoc.requestor.requested_contexts[0].context_id == 1
        assoc.release()
        assert not assoc.is_established
        assert assoc.is_released

        scp.shutdown()

    def test_associate_context_raises(self):
        """Test that AE.associate raises exception if no requested contexts"""
        self.ae = ae = AE()
        with pytest.raises(RuntimeError):
            assoc = ae.associate("localhost", 11112)


class TestAEGoodTimeoutSetters:
    def test_acse_timeout(self):
        """Check AE ACSE timeout change produces good value"""
        ae = AE()
        assert ae.acse_timeout == 30
        ae.acse_timeout = None
        assert ae.acse_timeout is None
        ae.acse_timeout = -100
        assert ae.acse_timeout == 30
        ae.acse_timeout = "a"
        assert ae.acse_timeout == 30
        ae.acse_timeout = 0
        assert ae.acse_timeout == 0
        ae.acse_timeout = 30
        assert ae.acse_timeout == 30

    def test_dimse_timeout(self):
        """Check AE DIMSE timeout change produces good value"""
        ae = AE()
        assert ae.dimse_timeout == 30
        ae.dimse_timeout = None
        assert ae.dimse_timeout is None
        ae.dimse_timeout = -100
        assert ae.dimse_timeout == 30
        ae.dimse_timeout = "a"
        assert ae.dimse_timeout == 30
        ae.dimse_timeout = 0
        assert ae.dimse_timeout == 0
        ae.dimse_timeout = 30
        assert ae.dimse_timeout == 30

    def test_network_timeout(self):
        """Check AE network timeout change produces good value"""
        ae = AE()
        assert ae.network_timeout == 60
        ae.network_timeout = None
        assert ae.network_timeout is None
        ae.network_timeout = -100
        assert ae.network_timeout == 60
        ae.network_timeout = "a"
        assert ae.network_timeout == 60
        ae.network_timeout = 0
        assert ae.network_timeout == 0
        ae.network_timeout = 30
        assert ae.network_timeout == 30

    def test_connection_timeout(self):
        """Check AE connection timeout change produces good value"""
        ae = AE()
        assert ae.connection_timeout is None
        ae.connection_timeout = None
        assert ae.connection_timeout is None
        ae.connection_timeout = -100
        assert ae.connection_timeout is None
        ae.connection_timeout = "a"
        assert ae.connection_timeout is None
        ae.connection_timeout = 0
        assert ae.connection_timeout is None
        ae.connection_timeout = 30
        assert ae.connection_timeout == 30

    def test_active_acse(self):
        """Test changing acse_timeout with active associations."""
        ae = AE()
        ae.add_supported_context("1.2.840.10008.1.1")
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context("1.2.840.10008.1.1")
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assert assoc.acse_timeout == 30
        ae.acse_timeout = 5
        assert assoc.acse_timeout == 5

        assoc.release()

        scp.shutdown()
        ae.shutdown()

    def test_active_dimse(self):
        """Test changing dimse_timeout with active associations."""
        ae = AE()
        ae.add_supported_context("1.2.840.10008.1.1")
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context("1.2.840.10008.1.1")
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assert assoc.dimse_timeout == 30
        ae.dimse_timeout = 5
        assert assoc.dimse_timeout == 5

        assoc.release()

        scp.shutdown()

    def test_active_network(self):
        """Test changing network_timeout with active associations."""
        ae = AE()
        ae.add_supported_context("1.2.840.10008.1.1")
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context("1.2.840.10008.1.1")
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assert assoc.network_timeout == 60
        ae.network_timeout = 5
        assert assoc.network_timeout == 5

        assoc.release()

        scp.shutdown()

    def test_active_connection(self):
        """Test changing connection_timeout with active associations."""
        ae = AE()
        ae.add_supported_context("1.2.840.10008.1.1")
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context("1.2.840.10008.1.1")
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assert assoc.connection_timeout is None
        ae.connection_timeout = 5
        assert assoc.connection_timeout == 5

        assoc.release()

        scp.shutdown()


class TestAEGoodAssociation:
    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_associate_establish_release(self):
        """Check SCU Association with SCP"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.release()
        assert not assoc.is_established
        assert assoc.is_released

        scp.shutdown()

    def test_associate_max_pdu(self):
        """Check Association has correct max PDUs on either end"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.maximum_pdu_size = 54321
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        scu_ae = AE()
        scu_ae.acse_timeout = 5
        scu_ae.dimse_timeout = 5
        scu_ae.network_timeout = 5
        scu_ae.add_requested_context(Verification)
        assoc = scu_ae.associate("localhost", 11112, max_pdu=12345)
        assert assoc.is_established

        assert scp.active_associations[0].acceptor.maximum_length == (54321)
        assert scp.active_associations[0].requestor.maximum_length == (12345)
        assert assoc.requestor.maximum_length == 12345
        assert assoc.acceptor.maximum_length == 54321
        assoc.release()

        time.sleep(0.1)
        assert scp.active_associations == []

        # Check 0 max pdu value - max PDU value maps to 0x10000 internally
        assoc = scu_ae.associate("localhost", 11112, max_pdu=0)
        assert assoc.requestor.maximum_length == 0
        assert scp.active_associations[0].requestor.maximum_length == 0

        assoc.release()

        scp.shutdown()

    def test_association_timeouts(self):
        """Check that the Association timeouts are being set correctly and
        work"""

        acse_delay = None
        dimse_delay = None

        def handle_echo(event):
            if dimse_delay:
                time.sleep(dimse_delay)

            return 0x0000

        def handle_acse_recv(event):
            if acse_delay:
                time.sleep(acse_delay)

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 0.5
        ae.add_supported_context(Verification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[
                (evt.EVT_ACSE_RECV, handle_acse_recv),
                (evt.EVT_C_ECHO, handle_echo),
            ],
        )

        scu_ae = AE()
        scu_ae.acse_timeout = 30
        scu_ae.dimse_timeout = 30
        scu_ae.network_timeout = 30
        scu_ae.add_requested_context(Verification)
        assoc = scu_ae.associate("localhost", 11112)
        assert assoc.is_established

        # Hit the network timeout
        time.sleep(1.0)
        assert assoc.is_aborted
        assert len(scp.active_associations) == 0

        ae.acse_timeout = None
        ae.dimse_timeout = None
        ae.network_timeout = None

        scu_ae.acse_timeout = 30
        scu_ae.dimse_timeout = 0

        dimse_delay = 1

        assoc = scu_ae.associate("localhost", 11112)
        assert assoc.is_established
        status = assoc.send_c_echo()
        time.sleep(1.5)
        assert assoc.is_aborted
        assert len(scp.active_associations) == 0

        # FIXME: If this is `0` we can process an ABORT primitive where
        # we expect an ASSOCIATION primitive.
        scu_ae.acse_timeout = 0.5
        scu_ae.dimse_timeout = 30

        acse_delay = 1

        assoc = scu_ae.associate("localhost", 11112)
        assert not assoc.is_established
        assert assoc.is_aborted
        time.sleep(1.5)
        assert len(scp.active_associations) == 0

        scu_ae.acse_timeout = 30
        # `0` is an invalid value
        scu_ae.connection_timeout = 0.5
        scu_ae.dimse_timeout = 30

        # The host exists and is routable, but there is a middlebox ignoring
        # the initial TCP SYN.
        assoc = scu_ae.associate("example.com", 11112)
        assert not assoc.is_established
        assert assoc.is_aborted
        assert len(scp.active_associations) == 0

        ae.acse_timeout = 21
        ae.dimse_timeout = 22
        scu_ae.acse_timeout = 31
        scu_ae.connection_timeout = None
        scu_ae.dimse_timeout = 32

        assoc = scu_ae.associate("localhost", 11112)
        assert assoc.is_established

        assert scp.active_associations[0].acse_timeout == 21
        assert scp.active_associations[0].dimse_timeout == 22
        assert assoc.acse_timeout == 31
        assert assoc.dimse_timeout == 32

        assoc.release()

        scp.shutdown()

    def test_connection_timeout(self, caplog):
        # * ACSE timeout does not start until connection timeout completes
        # * Logs indicate that we hit the timeout case
        scu_ae = AE()
        scu_ae.acse_timeout = 0.5
        scu_ae.connection_timeout = 1
        scu_ae.add_requested_context(Verification)
        with caplog.at_level(logging.ERROR, logger="pynetdicom"):
            assoc = scu_ae.associate(
                "8.8.8.8",
                11112,
                bind_address=("", 0),
            )
            assert not assoc.is_established
            assert assoc.is_aborted
            msgs = [
                "TCP Initialisation Error: timed out",
                "TCP Initialisation Error: [Errno -2] Name or service not known",
                # "TCP Initialisation Error: [Errno 113] No route to host",
            ]
            assert len([m for m in msgs if m in caplog.text]) == 1

    def test_select_timeout_okay(self):
        """Test that using start works OK with timeout."""
        # Multiple release/association in a sort time causes an OSError as
        # the port is still in use due to the use of select.select() with
        # a timeout. Fixed by using socket.shutdown in stop()
        for ii in range(3):
            self.ae = ae = AE()
            ae.acse_timeout = 5
            ae.dimse_timeout = 5
            ae.network_timeout = 5
            ae.add_supported_context(Verification)
            scp = ae.start_server(("localhost", 11112), block=False)

            ae.add_requested_context(Verification)
            assoc = ae.associate("localhost", 11112)
            assert assoc.is_established
            assoc.release()
            assert assoc.is_released
            assert not assoc.is_established

            scp.shutdown()

    def test_aet_bytes_deprecation(self):
        """Test warning if using bytes to set an AE title."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        server = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)
        msg = (
            r"The use of bytes with 'ae_title' is deprecated, use an ASCII "
            r"str instead"
        )
        with pytest.warns(DeprecationWarning, match=msg):
            assoc = ae.associate("localhost", 11112, ae_title=b"BADAE2")
            assert assoc.acceptor.ae_title == "BADAE2"
            assert assoc.requestor.ae_title == "PYNETDICOM"

        server.shutdown()


class TestAEBadAssociation:
    def test_raise(self):
        """Test bad associate call"""
        ae = AE()
        ae.add_requested_context(Verification)

        with pytest.raises(TypeError):
            ae.associate(1112, 11112)
        with pytest.raises(TypeError):
            ae.associate("localhost", "1.2.3.4")

    def test_invalid_ae_title(self):
        """Test invalid AE.ae_title"""
        ae = AE()
        ae.add_requested_context(Verification)
        msg = r"Invalid 'ae_title' value - must not consist entirely of spaces"
        with pytest.raises(ValueError, match=msg):
            ae.associate("localhost", 11112, ae_title="                ")

        msg = (
            r"Invalid 'ae_title' value '\u200b5' "
            r"- must only contain ASCII characters"
        )
        with pytest.raises(ValueError, match=msg):
            aet = b"\xe2\x80\x8b\x35".decode("utf8")
            ae.associate("localhost", 11112, ae_title=aet)

        msg = (
            r"Invalid 'ae_title' value '1234567890ABCDEFG' "
            r"- must not exceed 16 characters"
        )
        with pytest.raises(ValueError, match=msg):
            ae.associate("localhost", 11112, ae_title="1234567890ABCDEFG")

        msg = r"Invalid 'ae_title' value - must not be an empty str"
        with pytest.raises(ValueError, match=msg):
            ae.associate("localhost", 11112, ae_title="")

        msg = (
            r"Invalid 'ae_title' value 'TEST\\ME' - must not contain control "
            r"characters or backslashes"
        )
        with pytest.raises(ValueError, match=msg):
            ae.associate("localhost", 11112, ae_title="TEST\\ME")

        msg = r"'ae_title' must be str, not 'int'"
        with pytest.raises(TypeError, match=msg):
            ae.associate("localhost", 11112, ae_title=12345)


class TestAEGoodMiscSetters:
    def setup_method(self):
        self.ae = None

    def teardown_method(self):
        if self.ae:
            self.ae.shutdown()

    def test_ae_title_good(self):
        """Check AE title change produces good value"""
        ae = AE()
        ae.ae_title = "     TEST     "
        assert ae.ae_title == "     TEST     "
        ae.ae_title = "            TEST"
        assert ae.ae_title == "            TEST"
        ae.ae_title = "a            TES"
        assert ae.ae_title == "a            TES"
        ae.ae_title = "a        TEST"
        assert ae.ae_title == "a        TEST"

    def test_aet_bytes_deprecation(self):
        """Test warning if using bytes to set an AE title."""
        msg = (
            r"The use of bytes with 'ae_title' is deprecated, use an ASCII "
            r"str instead"
        )
        with pytest.warns(DeprecationWarning, match=msg):
            ae = AE(b"BADAE")
            assert ae.ae_title == "BADAE"

    def test_implementation(self):
        """Check the implementation version name and class UID setters"""
        ae = AE()
        ae.implementation_version_name = None
        assert ae.implementation_version_name is None
        ae.implementation_class_uid = "1.2.3"
        assert ae.implementation_class_uid == "1.2.3"

    def test_max_assoc_good(self):
        """Check AE maximum association change produces good value"""
        ae = AE()
        ae.maximum_associations = -10
        assert ae.maximum_associations == 1
        ae.maximum_associations = ["a"]
        assert ae.maximum_associations == 1
        ae.maximum_associations = "10"
        assert ae.maximum_associations == 1
        ae.maximum_associations = 0
        assert ae.maximum_associations == 1
        ae.maximum_associations = 5
        assert ae.maximum_associations == 5

    def test_max_pdu_good(self):
        """Check AE maximum pdu size change produces good value"""
        ae = AE()
        ae.maximum_pdu_size = -10
        assert ae.maximum_pdu_size == 16382
        ae.maximum_pdu_size = 0
        assert ae.maximum_pdu_size == 0
        ae.maximum_pdu_size = 5000
        assert ae.maximum_pdu_size == 5000

    def test_require_calling_aet(self):
        """Test AE.require_calling_aet"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established

        ae.require_calling_aet = ["MYAE"]
        assert ae.require_calling_aet == ["MYAE"]
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_rejected

        ae.require_calling_aet = ["PYNETDICOM"]
        assert ae.require_calling_aet == ["PYNETDICOM"]
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assoc.release()

        msg = r"Invalid 'require_calling_aet' value - must not be an empty str"
        with pytest.raises(ValueError, match=msg):
            ae.require_calling_aet = [""]
        assert ae.require_calling_aet == ["PYNETDICOM"]
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assoc.release()

        scp.shutdown()

    def test_aec_bytes_deprecation(self):
        """Test warning if using bytes to set an AE title."""
        ae = AE()
        msg = (
            r"The use of a list of bytes with 'require_calling_aet' is "
            r"deprecated, use a list of ASCII str instead"
        )
        with pytest.warns(DeprecationWarning, match=msg):
            ae.require_calling_aet = [b"BADAE", "GOODAE"]

        assert ae.require_calling_aet == ["BADAE", "GOODAE"]

    def test_require_called_aet(self):
        """Test AE.require_called_aet"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established

        ae.require_called_aet = True
        assert ae.require_called_aet is True
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_rejected

        assoc = ae.associate("localhost", 11112, ae_title="PYNETDICOM")
        assert assoc.is_established
        assoc.release()

        scp.shutdown()

    def test_req_calling_aet(self):
        """Check AE require calling aet change produces good value"""
        ae = AE()
        ae.require_calling_aet = ["10", "asdf"]
        assert ae.require_calling_aet == ["10", "asdf"]

    def test_req_called_aet(self):
        """Check AE require called aet change produces good value"""
        ae = AE()
        assert ae.require_called_aet is False
        ae.require_called_aet = True
        assert ae.require_called_aet is True
        ae.require_called_aet = False
        assert ae.require_called_aet is False

    def test_string_output(self):
        """Test string output"""
        ae = AE()
        ae.add_requested_context(Verification)
        ae.require_calling_aet = ["something"]
        ae.require_called_aet = True
        assert "Explicit VR" in ae.__str__()
        assert "Verification" in ae.__str__()
        assert "0/10" in ae.__str__()
        assert "something" in ae.__str__()
        assert "Require called AE title: True" in ae.__str__()
        ae.supported_contexts = StoragePresentationContexts
        assert "CT Image" in ae.__str__()

        ae = AE()
        ae.add_requested_context(Verification)
        assert "None" in ae.__str__()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established
        assert assoc.is_established
        assert "Explicit VR" in ae.__str__()
        assert "Peer" in ae.__str__()

        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established

        scp.shutdown()

    def test_init_implementation_class(self):
        """Test the default implementation class uid"""
        ae = AE()
        assert ae.implementation_class_uid == PYNETDICOM_IMPLEMENTATION_UID

    def test_init_implementation_version(self):
        """Test the default implementation version name"""
        ae = AE()
        assert ae.implementation_version_name == PYNETDICOM_IMPLEMENTATION_VERSION

    def test_implementation_version(self):
        """Test implementation_version_name"""
        ae = AE()
        ae.implementation_version_name = None
        assert ae.implementation_version_name is None
        ae.implementation_version_name = "  "
        assert ae.implementation_version_name == "  "

        msg = "'implementation_version_name' must be str or None, not 'int'"
        with pytest.raises(TypeError, match=msg):
            ae.implementation_version_name = 1234

        msg = "Invalid 'implementation_version_name' value - must not be an empty str"
        with pytest.raises(ValueError, match=msg):
            ae.implementation_version_name = ""

        assert ae.implementation_version_name == "  "

    def test_implementation_class(self):
        """Test implementation_class_uid"""
        ae = AE()
        ae.implementation_class_uid = "12.3.4"
        assert isinstance(ae.implementation_class_uid, UID)
        assert ae.implementation_class_uid == UID("12.3.4")

        msg = (
            r"'implementation_class_uid' must be str, bytes or UID, not " r"'NoneType'"
        )
        with pytest.raises(TypeError, match=msg):
            ae.implementation_class_uid = None

        assert ae.implementation_class_uid == UID("12.3.4")

        msg = r"Invalid 'implementation_class_uid' value - must not be an " r"empty str"
        with pytest.raises(ValueError, match=msg):
            ae.implementation_class_uid = ""

        msg = r"Invalid 'implementation_class_uid' value '1.2.04'"
        with pytest.raises(ValueError, match=msg):
            ae.implementation_class_uid = "1.2.04"

        assert ae.implementation_class_uid == UID("12.3.4")


class TestAEBadInitialisation:
    def test_invalid_ae_title(self):
        """Test invalid AE.ae_title"""
        msg = r"Invalid 'ae_title' value - must not consist entirely of spaces"
        with pytest.raises(ValueError, match=msg):
            AE(ae_title="                ")

        msg = (
            r"Invalid 'ae_title' value '\u200b5' "
            r"- must only contain ASCII characters"
        )
        with pytest.raises(ValueError, match=msg):
            AE(ae_title=b"\xe2\x80\x8b\x35".decode("utf8"))

        msg = (
            r"Invalid 'ae_title' value '1234567890ABCDEFG' "
            r"- must not exceed 16 characters"
        )
        with pytest.raises(ValueError, match=msg):
            AE(ae_title="1234567890ABCDEFG")

        msg = r"Invalid 'ae_title' value - must not be an empty str"
        with pytest.raises(ValueError, match=msg):
            AE(ae_title="")

        msg = (
            r"Invalid 'ae_title' value 'TEST\\ME' - must not contain control "
            r"characters or backslashes"
        )
        with pytest.raises(ValueError, match=msg):
            AE(ae_title="TEST\\ME")

        msg = r"'ae_title' must be str, not 'NoneType'"
        with pytest.raises(TypeError, match=msg):
            AE(ae_title=None)


class TestAE_GoodExit:
    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_ae_release_assoc(self):
        """Association releases OK"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)

        # Test N associate/release cycles
        for ii in range(5):
            assoc = ae.associate("localhost", 11112)
            assert assoc.is_established
            assoc.release()
            assert not assoc.is_established
            assert not assoc.is_aborted
            assert assoc.is_released
            assert not assoc.is_rejected

        scp.shutdown()

    def test_ae_aborts_assoc(self):
        """Association aborts OK"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(Verification)

        # Test N associate/abort cycles
        for ii in range(5):
            assoc = ae.associate("localhost", 11112)
            assert assoc.is_established
            assoc.abort()
            assert not assoc.is_established
            assert assoc.is_aborted
            assert not assoc.is_released
            assert not assoc.is_rejected

        scp.shutdown()


class TestAESupportedPresentationContexts:
    """Tests for AE's presentation contexts when acting as an SCP"""

    def setup_method(self):
        self.ae = AE()

    def test_add_supported_context_str(self):
        """Tests for AE.add_supported_context using str."""
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.context_id is None

    def test_add_supported_context_sop_class(self):
        """Tests for AE.add_supported_context using SOPClass."""
        self.ae.add_supported_context(RTImageStorage)

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_supported_context_uid(self):
        """Tests for AE.add_supported_context using UID."""
        self.ae.add_supported_context(UID("1.2.840.10008.1.1"))

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_supported_context_duplicate(self):
        """Tests for AE.add_supported_context using a duplicate UID."""
        self.ae.add_supported_context(UID("1.2.840.10008.1.1"))
        self.ae.add_supported_context(UID("1.2.840.10008.1.1"))

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_supported_context_transfer_single(self):
        """Test adding a single transfer syntax without a list"""
        self.ae.add_supported_context("1.2", "1.3")

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2"
        assert contexts[0].transfer_syntax == ["1.3"]

        self.ae.add_supported_context("1.2", UID("1.4"))

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2"
        assert contexts[0].transfer_syntax == ["1.3", "1.4"]

    def test_add_supported_context_duplicate_transfer(self):
        """Test adding duplicate transfer syntaxes."""
        self.ae.add_supported_context("1.2", ["1.3", "1.3"])

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2"
        assert contexts[0].transfer_syntax == ["1.3"]

        self.ae.supported_contexts = []
        self.ae.add_supported_context("1.2.840.10008.1.1")
        self.ae.add_supported_context("1.2.840.10008.1.1")

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.supported_contexts = []
        self.ae.add_supported_context("1.2.840.10008.1.1")
        self.ae.add_supported_context(
            "1.2.840.10008.1.1", [DEFAULT_TRANSFER_SYNTAXES[0]]
        )

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_supported_context_duplicate_multi(self):
        """Tests for AE.add_supported_context using a duplicate UID."""
        self.ae.add_supported_context(
            "1.2.840.10008.1.1", [DEFAULT_TRANSFER_SYNTAXES[0]]
        )
        self.ae.add_supported_context(
            "1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES[1:]
        )

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_supported_context_private_abs(self):
        """Test AE.add_supported_context with a private abstract syntax"""
        self.ae.add_supported_context("1.2.3.4")

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.3.4"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_supported_context_private_tran(self):
        """Test AE.add_supported_context with a private transfer syntax"""
        self.ae.add_supported_context("1.2.3.4", ["1.2.3", "1.2.840.10008.1.1"])

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.3.4"
        assert contexts[0].transfer_syntax == ["1.2.3", "1.2.840.10008.1.1"]

    def test_add_supported_context_more_128(self):
        """Test adding more than 128 presentation contexts"""
        for ii in range(300):
            self.ae.add_supported_context(str(ii))

        contexts = self.ae.supported_contexts
        assert len(contexts) == 300

    def test_supported_contexts_setter(self):
        """Test the AE.supported_contexts property setter."""
        context = build_context("1.2.840.10008.1.1")
        self.ae.supported_contexts = [context]

        contexts = self.ae.supported_contexts
        assert len(contexts) == 1
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.context_id is None

    def test_supported_contexts_empty(self):
        """Test the setting supported_contexts to an empty list."""
        context = build_context("1.2.840.10008.1.1")
        self.ae.supported_contexts = [context]
        assert len(self.ae.supported_contexts) == 1

        self.ae.supported_contexts = []
        assert len(self.ae.supported_contexts) == 0

    def test_supported_contexts_setter_raises(self):
        """Test the AE.supported_contexts property raises if not context."""
        with pytest.raises(ValueError):
            self.ae.supported_contexts = ["1.2.3"]

    def test_supported_contexts_sorted(self):
        """Test that the supported_contexts returns contexts in order."""
        self.ae.add_supported_context("1.2.3.4")
        self.ae.add_supported_context("1.2.3.5")

        asyntaxes = [cntx.abstract_syntax for cntx in self.ae.supported_contexts]
        assert asyntaxes == ["1.2.3.4", "1.2.3.5"]

        self.ae.add_supported_context("0.1.2.3")
        self.ae.add_supported_context("2.1.2.3")
        asyntaxes = [cntx.abstract_syntax for cntx in self.ae.supported_contexts]
        assert asyntaxes == ["0.1.2.3", "1.2.3.4", "1.2.3.5", "2.1.2.3"]

    def test_supported_contexts_more_128(self):
        """Test setting supported_contexts with more than 128 contexts."""
        contexts = []
        for ii in range(300):
            contexts.append(build_context(str(ii)))

        self.ae.supported_contexts = contexts
        assert len(self.ae.supported_contexts) == 300

    def test_remove_supported_context_str(self):
        """Tests for AE.remove_supported_context using str."""
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.remove_supported_context("1.2.840.10008.1.1")
        assert len(self.ae.supported_contexts) == 0

        # Test multiple
        self.ae.add_supported_context("1.2.840.10008.1.1")
        self.ae.add_supported_context("1.2.840.10008.1.4", ["1.2.3.4"])

        assert len(self.ae.supported_contexts) == 2
        self.ae.remove_supported_context("1.2.840.10008.1.1")
        assert len(self.ae.supported_contexts) == 1

        for context in self.ae.supported_contexts:
            assert context.abstract_syntax != "1.2.840.10008.1.1"

    def test_remove_supported_context_uid(self):
        """Tests for AE.remove_supported_context using UID."""
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.remove_supported_context(UID("1.2.840.10008.1.1"))
        assert len(self.ae.supported_contexts) == 0

    def test_remove_supported_context_sop_class(self):
        """Tests for AE.remove_supported_context using SOPClass."""
        self.ae.add_supported_context(RTImageStorage)

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.remove_supported_context(RTImageStorage)
        assert len(self.ae.supported_contexts) == 0

    def test_remove_supported_context_default(self):
        """Tests for AE.remove_supported_context with default transfers."""
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_supported_context("1.2.840.10008.1.1")
        assert len(self.ae.supported_contexts) == 0

    def test_remove_supported_context_single_transfer(self):
        """Tests for AE.remove_supported_context with single transfer."""
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_supported_context(
            "1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES[0]
        )
        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]

    def test_remove_supported_context_partial(self):
        """Tests for AE.remove_supported_context with partial transfers."""
        # Test singular
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_supported_context("1.2.840.10008.1.1", ["1.2.840.10008.1.2"])
        assert len(self.ae.supported_contexts) == 1
        context = self.ae.supported_contexts[0]
        assert len(context.transfer_syntax) == 3
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]
        assert context.abstract_syntax == "1.2.840.10008.1.1"

        # Test multiple
        self.ae.add_supported_context("1.2.840.10008.1.1")
        self.ae.add_supported_context(RTImageStorage)

        self.ae.remove_supported_context("1.2.840.10008.1.1", ["1.2.840.10008.1.2"])
        assert len(self.ae.supported_contexts) == 2
        context = self.ae.supported_contexts[0]
        assert len(context.transfer_syntax) == 3
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]
        assert context.abstract_syntax == "1.2.840.10008.1.1"

        assert (
            self.ae.supported_contexts[1].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        )

    def test_remove_supported_context_all(self):
        """Tests for AE.remove_supported_context with all transfers."""
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        # Test singular
        self.ae.remove_supported_context("1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES)
        assert len(self.ae.supported_contexts) == 0

        # Test multiple
        self.ae.add_supported_context("1.2.840.10008.1.1")
        self.ae.add_supported_context(RTImageStorage)

        self.ae.remove_supported_context("1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES)

        context = self.ae.supported_contexts[0]
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"

    def test_remove_supported_context_all_plus(self):
        """Test remove_supported_context with extra transfers"""
        tsyntax = DEFAULT_TRANSFER_SYNTAXES[:]
        tsyntax.append("1.2.3")
        self.ae.add_supported_context("1.2.840.10008.1.1")

        context = self.ae.supported_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_supported_context("1.2.840.10008.1.1", tsyntax)
        assert len(self.ae.supported_contexts) == 0

    def test_scu_role(self):
        """Test add_supported_context with scu_role parameter."""
        self.ae.add_supported_context("1.2.3")
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.supported_context = []
        self.ae.add_supported_context("1.2.3", scu_role=None)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.supported_context = []
        self.ae.add_supported_context("1.2.3", scu_role=True)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is True
        assert context.scp_role is None

        self.ae.supported_context = []
        self.ae.add_supported_context("1.2.3", scu_role=False)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is False
        assert context.scp_role is None

    def test_scu_role_update(self):
        """Test updating add_supported_context with scu_role parameter."""
        self.ae.add_supported_context("1.2.3")
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.add_supported_context("1.2.3", scu_role=None)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.add_supported_context("1.2.3", scu_role=True)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is True
        assert context.scp_role is None

        self.ae.add_supported_context("1.2.3", scu_role=False)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is False
        assert context.scp_role is None

    def test_scu_role_raises(self):
        """Test add_supported_context raises if scu_role wrong type."""
        with pytest.raises(TypeError, match=""):
            self.ae.add_supported_context("1.2.3", scu_role="abc")

        assert self.ae.supported_contexts == []

    def test_scp_role(self):
        """Test add_supported_context with scu_role parameter."""
        self.ae.add_supported_context("1.2.3")
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.supported_context = []
        self.ae.add_supported_context("1.2.3", scp_role=None)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.supported_context = []
        self.ae.add_supported_context("1.2.3", scp_role=True)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is True

        self.ae.supported_context = []
        self.ae.add_supported_context("1.2.3", scp_role=False)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is False

    def test_scp_role_update(self):
        """Test updating add_supported_context with scp_role parameter."""
        self.ae.add_supported_context("1.2.3")
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.add_supported_context("1.2.3", scp_role=None)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is None

        self.ae.add_supported_context("1.2.3", scp_role=True)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is True

        self.ae.add_supported_context("1.2.3", scp_role=False)
        context = self.ae.supported_contexts[0]
        assert context.scu_role is None
        assert context.scp_role is False

    def test_scp_role_raises(self):
        """Test add_supported_context raises if scp_role wrong type."""
        with pytest.raises(TypeError, match=""):
            self.ae.add_supported_context("1.2.3", scp_role="abc")

        assert self.ae.supported_contexts == []


class TestAERequestedPresentationContexts:
    """Tests for AE's presentation contexts when acting as an SCU"""

    def setup_method(self):
        self.ae = AE()

    def test_add_requested_context_str(self):
        """Tests for AE.add_requested_context using str."""
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.context_id is None

    def test_add_requested_context_sop_class(self):
        """Tests for AE.add_requested_context using SOPClass."""
        self.ae.add_requested_context(RTImageStorage)

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_requested_context_uid(self):
        """Tests for AE.add_requested_context using UID."""
        self.ae.add_requested_context(UID("1.2.840.10008.1.1"))

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_requested_context_duplicate(self):
        """Test AE.add_requested_context using a duplicate UID."""
        self.ae.add_requested_context(UID("1.2.840.10008.1.1"))
        self.ae.add_requested_context(UID("1.2.840.10008.1.1"))

        contexts = self.ae.requested_contexts
        assert len(contexts) == 2
        assert contexts[0].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert contexts[1].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[1].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_requested_context_duplicate_multi(self):
        """Tests for AE.add_requested_context using a duplicate UID."""
        self.ae.add_requested_context(
            "1.2.840.10008.1.1", [DEFAULT_TRANSFER_SYNTAXES[0]]
        )
        self.ae.add_requested_context(
            "1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES[1:]
        )

        contexts = self.ae.requested_contexts
        assert len(contexts) == 2
        assert contexts[0].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[0].transfer_syntax == [DEFAULT_TRANSFER_SYNTAXES[0]]
        assert contexts[1].abstract_syntax == "1.2.840.10008.1.1"
        assert contexts[1].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]

    def test_add_supported_context_transfer_single(self):
        """Test adding a single transfer syntax without a list"""
        self.ae.add_requested_context("1.2", "1.3")

        contexts = self.ae.requested_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2"
        assert contexts[0].transfer_syntax == ["1.3"]

        self.ae.add_requested_context("1.2", UID("1.4"))

        contexts = self.ae.requested_contexts
        assert len(contexts) == 2
        assert contexts[1].abstract_syntax == "1.2"
        assert contexts[1].transfer_syntax == ["1.4"]

    def test_add_requested_context_duplicate_transfer(self):
        """Test add_requested_context using duplicate transfer syntaxes"""
        self.ae.add_requested_context("1.2", ["1.3", "1.3"])
        contexts = self.ae.requested_contexts
        assert contexts[0].transfer_syntax == ["1.3"]

    def test_add_requested_context_private_abs(self):
        """Test AE.add_requested_context with a private abstract syntax"""
        self.ae.add_requested_context("1.2.3.4")

        contexts = self.ae.requested_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.3.4"
        assert contexts[0].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_add_requested_context_private_tran(self):
        """Test AE.add_requested_context with a private transfer syntax"""
        self.ae.add_requested_context("1.2.3.4", ["1.2.3", "1.2.840.10008.1.1"])

        contexts = self.ae.requested_contexts
        assert len(contexts) == 1
        assert contexts[0].abstract_syntax == "1.2.3.4"
        assert contexts[0].transfer_syntax == ["1.2.3", "1.2.840.10008.1.1"]

    def test_add_requested_context_more_128_raises(self):
        """Test adding more than 128 presentation contexts"""
        for ii in range(128):
            self.ae.add_requested_context(str(ii))

        assert len(self.ae.requested_contexts) == 128

        with pytest.raises(ValueError):
            self.ae.add_requested_context("129")

        assert len(self.ae.requested_contexts) == 128

    def test_requested_contexts_setter(self):
        """Test the AE.requested_contexts property setter."""
        context = build_context("1.2.840.10008.1.1")
        self.ae.requested_contexts = [context]

        contexts = self.ae.requested_contexts
        assert len(contexts) == 1
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.context_id is None

    def test_requested_contexts_empty(self):
        """Test the setting requested_contexts to an empty list."""
        context = build_context("1.2.840.10008.1.1")
        self.ae.requested_contexts = [context]
        assert len(self.ae.requested_contexts) == 1

        self.ae.requested_contexts = []
        assert len(self.ae.requested_contexts) == 0

    def test_requested_contexts_setter_raises(self):
        """Test the AE.requested_contexts property raises if not context."""
        with pytest.raises(ValueError):
            self.ae.requested_contexts = ["1.2.3"]

    def test_requested_contexts_not_sorted(self):
        """Test that requested_contexts returns contexts in supplied order."""
        self.ae.add_requested_context("1.2.3.4")
        self.ae.add_requested_context("1.2.3.5")

        asyntaxes = [cntx.abstract_syntax for cntx in self.ae.requested_contexts]
        assert asyntaxes == ["1.2.3.4", "1.2.3.5"]

        self.ae.add_requested_context("0.1.2.3")
        self.ae.add_requested_context("2.1.2.3")
        asyntaxes = [cntx.abstract_syntax for cntx in self.ae.requested_contexts]
        assert asyntaxes == ["1.2.3.4", "1.2.3.5", "0.1.2.3", "2.1.2.3"]

    def test_requested_contexts_more_128(self):
        """Test setting requested_contexts with more than 128 contexts."""
        contexts = []
        for ii in range(128):
            contexts.append(build_context(str(ii)))

        self.ae.requested_contexts = contexts
        assert len(self.ae.requested_contexts) == 128

        contexts.append(build_context("129"))

        with pytest.raises(ValueError):
            self.ae.requested_contexts = contexts

    def test_remove_requested_context_str(self):
        """Tests for AE.remove_requested_context using str."""
        # Test singular
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.remove_requested_context("1.2.840.10008.1.1")
        assert len(self.ae.requested_contexts) == 0

        # Test multiple
        self.ae.add_requested_context("1.2.840.10008.1.1")
        self.ae.add_requested_context("1.2.840.10008.1.1", ["1.2.3.4"])
        self.ae.add_requested_context("1.2.840.10008.1.4", ["1.2.3.4"])

        assert len(self.ae.requested_contexts) == 3
        self.ae.remove_requested_context("1.2.840.10008.1.1")
        assert len(self.ae.requested_contexts) == 1

        for context in self.ae.requested_contexts:
            assert context.abstract_syntax != "1.2.840.10008.1.1"

    def test_remove_requested_context_uid(self):
        """Tests for AE.remove_requested_context using UID."""
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.remove_requested_context(UID("1.2.840.10008.1.1"))
        assert len(self.ae.requested_contexts) == 0

    def test_remove_requested_context_sop_class(self):
        """Tests for AE.remove_requested_context using SOPClass."""
        self.ae.add_requested_context(RTImageStorage)

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

        self.ae.remove_requested_context(RTImageStorage)
        assert len(self.ae.requested_contexts) == 0

    def test_remove_requested_context_default(self):
        """Tests for AE.remove_requested_context with default transfers."""
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_requested_context("1.2.840.10008.1.1")
        assert len(self.ae.requested_contexts) == 0

    def test_remove_requested_context_single(self):
        """Tests for AE.remove_requested_context with single transfer."""
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_requested_context(
            "1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES[0]
        )
        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]

    def test_remove_requested_context_partial(self):
        """Tests for AE.remove_supported_context with partial transfers."""
        # Test singular
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_requested_context("1.2.840.10008.1.1", ["1.2.840.10008.1.2"])
        assert len(self.ae.requested_contexts) == 1
        context = self.ae.requested_contexts[0]
        assert len(context.transfer_syntax) == 3
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]
        assert context.abstract_syntax == "1.2.840.10008.1.1"

        self.ae.remove_requested_context("1.2.840.10008.1.1")
        assert len(self.ae.requested_contexts) == 0

        # Test multiple
        self.ae.add_requested_context("1.2.840.10008.1.1")
        self.ae.add_requested_context(RTImageStorage)
        self.ae.add_requested_context("1.2.840.10008.1.1", ["1.2.3.4"])

        self.ae.remove_requested_context("1.2.840.10008.1.1", ["1.2.840.10008.1.2"])
        assert len(self.ae.requested_contexts) == 3
        context = self.ae.requested_contexts[0]
        assert len(context.transfer_syntax) == 3
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES[1:]
        assert context.abstract_syntax == "1.2.840.10008.1.1"

        assert (
            self.ae.requested_contexts[1].transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        )
        assert self.ae.requested_contexts[2].transfer_syntax == ["1.2.3.4"]
        assert self.ae.requested_contexts[2].abstract_syntax == "1.2.840.10008.1.1"

        self.ae.remove_requested_context("1.2.840.10008.1.1")
        assert len(self.ae.requested_contexts) == 1
        assert (
            self.ae.requested_contexts[0].abstract_syntax
            == "1.2.840.10008.5.1.4.1.1.481.1"
        )

    def test_remove_requested_context_all(self):
        """Tests for AE.remove_requested_context with all transfers."""
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        # Test singular
        self.ae.remove_requested_context("1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES)
        assert len(self.ae.requested_contexts) == 0

        # Test multiple
        self.ae.add_requested_context(
            "1.2.840.10008.1.1", [DEFAULT_TRANSFER_SYNTAXES[0]]
        )
        self.ae.add_requested_context(
            "1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES[1:]
        )
        self.ae.add_requested_context(RTImageStorage)

        self.ae.remove_requested_context("1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES)

        assert len(self.ae.requested_contexts) == 1
        context = self.ae.requested_contexts[0]
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"

    def test_remove_requested_context_all_plus(self):
        """Test remove_requested_context with extra transfers"""
        tsyntax = DEFAULT_TRANSFER_SYNTAXES[:]
        tsyntax.append("1.2.3")
        # Test singular
        self.ae.add_requested_context("1.2.840.10008.1.1")

        context = self.ae.requested_contexts[0]
        assert context.abstract_syntax == "1.2.840.10008.1.1"
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert len(context.transfer_syntax) == 4

        self.ae.remove_requested_context("1.2.840.10008.1.1", tsyntax)
        assert len(self.ae.requested_contexts) == 0

        # Test multiple
        self.ae.add_requested_context(
            "1.2.840.10008.1.1", [DEFAULT_TRANSFER_SYNTAXES[0]]
        )
        self.ae.add_requested_context(
            "1.2.840.10008.1.1", DEFAULT_TRANSFER_SYNTAXES[1:]
        )
        self.ae.add_requested_context(RTImageStorage)

        self.ae.remove_requested_context("1.2.840.10008.1.1", tsyntax)

        assert len(self.ae.requested_contexts) == 1
        context = self.ae.requested_contexts[0]
        assert context.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES
        assert context.abstract_syntax == "1.2.840.10008.5.1.4.1.1.481.1"
